import colorsys
from jaxtyping import UInt8
import numpy as np
import pure_cv as vc
import cv2

def mod_color(i: int, n: int, lightness: float = 0.5, saturation: float = 0.8) -> tuple[int, int, int]:
    """Color per every `i mod n`, evenly spaced around the hue circle"""
    hue = i / n
    rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
    return tuple(int(x*255) for x in rgb) # type: ignore

def black(width: int, height: int) -> UInt8[np.ndarray, 'height width 3']:
    return np.zeros((height, width, 3), dtype=np.uint8)

def blend_texture(img: vc.Img, texture: vc.Img, alpha=0.7) -> vc.Img:
  """Alpha blend a texture on top of an image"""
  h, w = img.shape[:2]
  texture = cv2.resize(texture, (w, h))
  return cv2.addWeighted(img.astype(float), alpha, texture.astype(float), 1 - alpha, 0).astype(np.uint8)


from typing_extensions import Unpack
from jaxtyping import Shaped
import numpy as np
import cv2 as cv
import pure_cv as vc

def cartesian2homogeneous(points: Shaped[np.ndarray, '* 2']) -> Shaped[np.ndarray, '* 3']:
  """Convert cartesian points to homogeneous points"""
  ones = np.ones((*points.shape[:-1], 1))
  return np.concatenate([points, ones], axis=-1)

def homogeneous2cartesian(points: Shaped[np.ndarray, '* 3']) -> Shaped[np.ndarray, '* 2']:
  """Convert homogeneous points to cartesian points"""
  return points[..., :-1] / points[..., -1:]

def warp(points: Shaped[np.ndarray, '* 2'], homography: vc.Mat3) -> Shaped[np.ndarray, '* 2']:
  """Warp points using a homography"""
  M = np.array(homography)
  h = cartesian2homogeneous(points)
  h2 = h @ M.T
  return homogeneous2cartesian(h2)

def pad(corners: vc.Corners, *, padx: float, pady: float) -> vc.Corners:
  """Pad perspective corners, by:
  1. Move the centroid to 0
  2. Scale by `[1+padx, 1+pady]`
  3. Move back to original position
  """
  cs = np.array(corners)
  centroid = cs.mean(axis=0)
  return (cs - centroid) * np.array([1+padx, 1+pady]) + centroid

# def homography(corners: vc.Corners):
#   """Compute homography to map corners to a rectangle. Returns `homography, (width, height)`"""
#   tl, tr, br, bl = np.array(corners)
#   w = int((np.linalg.norm(tl-tr) + np.linalg.norm(bl-br)) / 2)
#   h = int((np.linalg.norm(tl-bl) + np.linalg.norm(tr-br)) / 2)
#   src = np.array([tl, tr, br, bl])
#   dst = np.array([[0, 0], [w, 0], [w, h], [0, h]])
#   return cv.getPerspectiveTransform(src.astype(np.float32), dst.astype(np.float32)), (w, h)

def homography(corners: vc.Corners, *, l=0., r=0., t=0., b=0.):
  """Compute homography to map corners to a rectangle. Returns `homography, (width, height)`
  - `l, r, t, b`: relative paddings to add around the rectangle
  """
  tl, tr, br, bl = np.array(corners)
  detected_w = int((np.linalg.norm(tl-tr) + np.linalg.norm(bl-br)) / 2)
  detected_h = int((np.linalg.norm(tl-bl) + np.linalg.norm(tr-br)) / 2)
  pad_left = int(detected_w*l); pad_right = int(detected_w*r)
  pad_top = int(detected_h*t); pad_bot = int(detected_h*b)
  w = int(detected_w + pad_left + pad_right)
  h = int(detected_h + pad_top + pad_bot)
  src = np.array([tl, tr, br, bl])
  dst = np.array([[pad_left, pad_top], [w-pad_right, pad_top], [w-pad_right, h-pad_bot], [pad_left, h-pad_bot]])
  return cv.getPerspectiveTransform(src.astype(np.float32), dst.astype(np.float32)), (w, h)

def correct(img: vc.Img, corners: vc.Corners, **pads: Unpack[vc.Pads]) -> vc.Img:
  """Correct image perspective (from absolute corners)
  - `l, r, t, b`: relative paddings to add around the rectangle (default to 0)
  """
  M, (w, h) = homography(corners, **pads)
  return cv.warpPerspective(img, M, (w, h)) 

def unwarp_contours(contours: vc.Contours, corners: vc.Corners) -> vc.Contours:
  """Re-coord contours w.r.t. to the original image
  - `corners`: perspective corners of the original image, yielding the corrected image
  - `contours`: contours w.r.t. the corrected image
  """
  M, _ = homography(corners)
  Minv = np.linalg.inv(M)
  return warp(contours, Minv)


from typing_extensions import Literal, Annotated, TypedDict, Sequence, TypeAlias
from nudantic import NdArray

Img2D = Annotated[NdArray, 'W H']
Img3D = Annotated[NdArray, 'W H C']
Img: TypeAlias = Img2D | Img3D | Annotated[NdArray, '*']

Contour = Annotated[NdArray, 'C 2']
"""`(x1, y1), ...`"""
Contours = Annotated[NdArray, 'N C 2']
"""`[(x1, y1), ...], ...]`"""

BBox = Annotated[NdArray, '2 2']
"""`(x1, y1), (x2, y2)`"""

BBoxes = Annotated[NdArray, 'N 2 2']
"""`[((x1, y1), (x2, y2)), ...]`"""

Vec2 = tuple[float, float] | Sequence[float] | Annotated[NdArray, '2']
Vertices = Annotated[NdArray, 'N 2'] | Sequence[Vec2]

Corners = Annotated[NdArray, '4 2']
"""`tl, tr, br, bl`"""

Segment = Annotated[NdArray, '1 4'] | Sequence[Sequence[float]]
Segments = Annotated[NdArray, 'N 1 4'] | Sequence[Segment]

Mat3 = Annotated[NdArray, '3 3']

class Rect(TypedDict):
  tl: Vec2
  size: Vec2

class Pads(TypedDict, total=False):
  l: float; r: float; t: float; b: float

Rotation = Literal[90, 180, -90]


from typing import Sequence, overload, Callable
from jaxtyping import Shaped
import cv2 as cv
import numpy as np
import ramda as R
import pure_cv as vc

@overload
def vertices(img: vc.Img, /, vertices: vc.Vertices, *, radius: int = 10, color: int | tuple[int, int, int] = (255, 0, 0)) -> vc.Img: ...
@overload
def vertices(*, vertices: vc.Vertices, radius: int = 10, color: int | tuple[int, int, int] = (255, 0, 0)) -> Callable[[vc.Img], vc.Img]: ...

@R.curry
def vertices( # type: ignore
  img: vc.Img, /, vertices: vc.Vertices, radius: int = 10, color: int | tuple[int, int, int] = (255, 0, 0)
) -> vc.Img:
    if isinstance(color, tuple):
        img = cv.cvtColor(img, cv.COLOR_GRAY2BGR) if len(np.shape(img)) == 2 else np.copy(img)
    output = img.copy()
    for [x, y] in vertices:
        cv.circle(output, (int(x), int(y)), radius=radius, color=color, thickness=-1) # type: ignore
    return output


@overload
def vlines(img: vc.Img, /, xs: list[float], *, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> vc.Img: ...
@overload
def vlines(*, xs: list[float], color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Callable[[vc.Img], vc.Img]: ...

@R.curry
def vlines( # type: ignore
    img: vc.Img, /, xs: list[float], color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3
) -> vc.Img:
    height = img.shape[0]
    ls = np.array([[[x, 0, x, height-1]] for x in xs], dtype=np.int32)
    return lines(img, ls, color=color, thickness=thickness)

@overload
def hlines(img: vc.Img, /, ys: list[float], *, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> vc.Img: ...
@overload
def hlines(*, ys: list[float], color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Callable[[vc.Img], vc.Img]: ...

@R.curry
def hlines( # type: ignore
    img: vc.Img, /, ys: list[float], color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3
) -> vc.Img:
    width = img.shape[1]
    ls = np.array([[[0, y, width-1, y]] for y in ys], dtype=np.int32)
    return lines(img, ls, color=color, thickness=thickness)



@overload
def lines(img: vc.Img, /, lines: vc.Segments, *, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> vc.Img: ...
@overload
def lines(*, lines: vc.Segments, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Callable[[vc.Img], vc.Img]: ...
@R.curry
def lines( # type: ignore
  img: vc.Img, /, lines: vc.Segments, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3
) -> vc.Img: # type: ignore
  out = img.copy()
  for line in lines:
    x1, y1, x2, y2 = line[0]
    cv.line(out, [x1, y1], [x2, y2], color=color, thickness=thickness) # type: ignore
  return out

@overload
def dashed_lines(
  img: vc.Img, /, lines: vc.Segments, *, 
  color: int | tuple[int, int, int] = (255, 0, 0), 
  thickness: int = 3, dash_length: int = 10, gap_length: int = 10
) -> vc.Img: ...

@overload
def dashed_lines(
  *, lines: vc.Segments, 
  color: int | tuple[int, int, int] = (255, 0, 0), 
  thickness: int = 3, dash_length: int = 10, gap_length: int = 10
) -> Callable[[vc.Img], vc.Img]: ...

@R.curry
def dashed_lines( # type: ignore
  img: vc.Img, /, lines: vc.Segments, 
  color: int | tuple[int, int, int] = (255, 0, 0), 
  thickness: int = 3, dash_length: int = 10, gap_length: int = 10
) -> vc.Img:
    out = img.copy()

    for line in lines:
      x1, y1, x2, y2 = line[0]
      pt1 = np.array([x1, y1])
      pt2 = np.array([x2, y2])
      length = np.linalg.norm(pt2 - pt1)

      # Calculate the number of dashes symmetrically
      pattern_length = dash_length + gap_length
      dashes = int(length // pattern_length)
      start_offset = (length - dashes * pattern_length) / 2
      
      for i in range(dashes):
        start = pt1 + (pt2 - pt1) * ((i * pattern_length + start_offset) / length)
        end = pt1 + (pt2 - pt1) * ((i * pattern_length + start_offset + dash_length) / length)
        cv.line(out, tuple(start.astype(int)), tuple(end.astype(int)), color=color, thickness=thickness) # type: ignore

    return out


@overload
def contours(img: vc.Img, /, contours: vc.Contours, *, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> vc.Img: ...
@overload
def contours(*, contours: vc.Contours, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Callable[[vc.Img], vc.Img]: ...
@R.curry
def contours( # type: ignore
  img: vc.Img, /, contours: vc.Contours, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3
) -> vc.Img:
  if isinstance(color, tuple):
    outimg = cv.cvtColor(img, cv.COLOR_GRAY2BGR) if len(np.shape(img)) == 2 else np.copy(img)
  else:
    outimg = np.copy(img) if len(np.shape(img)) == 2 else cv.cvtColor(img, cv.COLOR_BGR2GRAY)
    
  return cv.drawContours(outimg, [np.array(c, dtype=np.int32) for c in contours], -1, color, thickness) # type: ignore


@overload
def gradient_contours(
  img: vc.Img, /, cnts: vc.Contours, *,
  color0=(255, 0, 0), color1=(0, 255, 0),
  cycle: int | None = None
) -> vc.Img: ...
@overload
def gradient_contours(
  cnts: vc.Contours, *,
  color0=(255, 0, 0), color1=(0, 255, 0),
  cycle: int | None = None
) -> Callable[[vc.Img], vc.Img]: ...
@R.curry
def gradient_contours( # type: ignore
  img: vc.Img, cnts: vc.Contours, *,
  color0=(255, 0, 0), color1=(0, 255, 0),
  cycle: int | None = None
) -> vc.Img:
  cycle = cycle or len(cnts)
  c0, c1 = np.array(color0), np.array(color1)
  img = img.copy()
  for i, cnt in enumerate(cnts):
    alpha = i % cycle / cycle
    color = (c0 * (1 - alpha) + c1 * alpha).astype(np.uint8).tolist()
    img = vc.draw.contours(img, np.array([cnt]), color=tuple(color))
  return img

@overload
def bboxes(img: vc.Img, /, bboxes: vc.BBoxes, *, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> vc.Img: ...
@overload
def bboxes(*, bboxes: vc.BBoxes, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3) -> Callable[[vc.Img], vc.Img]: ...
@R.curry
def bboxes( # type: ignore
  img: vc.Img, /, bboxes: vc.BBoxes, color: int | tuple[int, int, int] = (255, 0, 0), thickness: int = 3
) -> vc.Img:
  cnts = vc.bbox2contour(bboxes)
  return contours(img, contours=cnts, color=color, thickness=thickness)
  


from typing_extensions import Unpack
import numpy as np
import pure_cv as vc

def bbox2contour(bboxes: vc.BBox | vc.BBoxes) -> vc.Contour | vc.Contours:
  """Convert a bounding box to a 4-corner contour. Supports vectorized operation"""
  X1 = bboxes[..., 0, 0] # (N,)
  Y1 = bboxes[..., 0, 1] # (N,)
  X2 = bboxes[..., 1, 0] # (N,)
  Y2 = bboxes[..., 1, 1] # (N,)
  TLs = np.stack([X1, Y1], axis=-1) # (N, 2)
  TRs = np.stack([X2, Y1], axis=-1) # (N, 2)
  BRs = np.stack([X2, Y2], axis=-1) # (N, 2)
  BLs = np.stack([X1, Y2], axis=-1) # (N, 2)
  return np.stack([TLs, TRs, BRs, BLs], axis=-2) # (N, 4, 2)

def contour2bbox(contours: vc.Contour | vc.Contours) -> vc.BBox | vc.BBoxes:
  """Convert a 4-corner contour to a bounding box. Supports vectorized operation"""
  xs = contours[..., 0] # (N,)
  ys = contours[..., 1] # (N,)
  X1 = np.min(xs, axis=-1) # (N,)
  Y1 = np.min(ys, axis=-1) # (N,)
  X2 = np.max(xs, axis=-1) # (N,)
  Y2 = np.max(ys, axis=-1) # (N,)
  TL = np.stack([X1, Y1], axis=-1) # (N, 2)
  BR = np.stack([X2, Y2], axis=-1) # (N, 2)
  return np.stack([TL, BR], axis=-2) # (N, 2, 2)

def roi(
  img: vc.Img, contour: vc.Contour, *,
  l = 0.1, r = 0.1, t = 0.15, b = 0.25
):
  """Extract a 4-corner contour by persective correcting it to a rectangle"""
  return vc.corners.correct(img, contour, l=l, r=r, t=t, b=b)



def pad_bboxes(bboxes: vc.BBoxes, *, l=0., r=0., t=0., b=0.) -> vc.BBoxes:
  """Pad bounding boxes, with padding relative to their shapes"""
  x0 = bboxes[..., 0, 0]
  y0 = bboxes[..., 0, 1]
  x1 = bboxes[..., 1, 0]
  y1 = bboxes[..., 1, 1]
  w = x1 - x0
  h = y1 - y0
  top = y0 - t*h
  left = x0 - l*w
  bottom = y1 + b*h
  right = x1 + r*w
  return np.stack([np.stack([left, top], axis=-1), np.stack([right, bottom], axis=-1)], axis=-2)

def extract_contours(img: vc.Img, contours: vc.Contours, **pads: Unpack[vc.Pads]) -> list[vc.Img]:
  """Extract contours from an image"""
  return [roi(img, contour, **pads) for contour in contours]


from .transform import homogeneous2cartesian, cartesian2homogeneous, \
  pad, homography, correct, warp, unwarp_contours

__all__ = [
  'homogeneous2cartesian', 'cartesian2homogeneous',
  'pad', 'homography', 'correct', 'warp', 'unwarp_contours',
]


from .types_ import Rotation, Img, Img2D, Img3D, Contour, Contours, BBox, Corners, Vec2, \
  Rect, Pads, Mat3, Vertices, BBoxes, Segment, Segments
from .opencv import imread, grayscale, threshold, dilate, erode, cvtColor, imdecode
try:
  from .plot import show
except ImportError:
  ...
from .encoding import b64encode, b64decode, encode, decode
from .misc import black, mod_color, blend_texture
from .rescaling import descale_h, rescale_h, rescale_w, descale_w, descale_max
from .rotation import rotate
from .boxes import bbox2contour, contour2bbox, roi, extract_contours, pad_bboxes
from . import draw, corners

__all__ = [
  'imread', 'grayscale', 'threshold', 'dilate', 'erode', 'cvtColor', 'imdecode', 
  'show', 'corners',
  'b64encode', 'b64decode', 'encode', 'decode', 'pad_bboxes',
  'bbox2contour', 'contour2bbox', 'roi', 'extract_contours',
  'mod_color', 'black', 'blend_texture',
  'descale_h', 'rescale_h', 'rescale_w', 'descale_w', 'descale_max',
  'rotate', 'Rotation',
  'draw', 'Pads', 'Mat3', 'Vertices', 'BBoxes', 'Segment', 'Segments',
  'Img', 'Img2D', 'Img3D', 'Contour', 'BBox', 'Contours', 'Corners', 'Vec2', 'Rect',
]
